// Copyright (c) 2005 David Drysdale.  See license.txt for license info
#include "board.h"
#include "boardgadget.h"
#include "constants.h"
#include "globals.h"
#include "utils.h"

void BoardShowNumbers(FormPtr form)
{
  Char countString[5];
  Int16 timeTaken;
  
  if (dontDraw) return;
  
  StrPrintF(countString, "%3d", board->minesLeftToMark);
  FrmHideObject(form, FrmGetObjectIndex(form, MinesLabel));
  FrmCopyLabel(form, MinesLabel, countString);
  FrmShowObject(form, FrmGetObjectIndex(form, MinesLabel));

  timeTaken = board->timeTaken / SysTicksPerSecond();
  StrPrintF(countString, "%3d", (timeTaken<=kTimeLimit) ? timeTaken : kTimeLimit);
  FrmHideObject(form, FrmGetObjectIndex(form, TimeLabel));
  FrmCopyLabel(form, TimeLabel, countString);
  FrmShowObject(form, FrmGetObjectIndex(form, TimeLabel));
}

void CellDraw(UInt16 x, UInt16 y, UInt16 xIx, UInt16 yIx)
{
  Coord x1, y1, x2, y2;
  MemHandle resH;
  BitmapType *bitmap;
  DmResID resourceID;
  UInt8 cell;

  if ((xIx >= board->width) || (yIx >= board->height)) return; 

  // Offset the coordinates; they are passed to us as gadget-relative, and we 
  // need them to be screen-relative
  x += mainDeltaX;
  y += mainDeltaY;

  // Draw the N line
  x1 = x; x2 = x + kCellSize - 1;
  y1 = y; y2 = y1;
  WinDrawLine(x1,y1,x2,y2);

  // Draw the W line
  x1 = x; x2 = x1;
  y1 = y; y2 = y + kCellSize - 1;
  WinDrawLine(x1,y1,x2,y2);

  cell = CELLVALUE(board, xIx, yIx);
  // DEBUG((T,"cell %d,%d val %x", xIx, yIx, cell));

  // Figure out the bitmap to use for the cell
  if (REVEALED(cell)) {
    if (MINE(cell)) {
      resourceID = MineBitmap;
    } else {
      if (FLAGGED(cell)) {
        resourceID = NotMineBitmap;
      } else {
        assert(NBRS(cell) <= 8);
        resourceID = Nbrs0Bitmap + NBRS(cell);
      }
    }
  } else if (FLAGGED(cell)) {
    resourceID = FlagBitmap;
  } else if (QUESTION(cell)) { 
    resourceID = QuestionBitmap;
  } else {
    // Normal unrevealed cell
    resourceID = NormalBitmap;
  }

  // Find and draw the relevant bitmap
  resH = DmGetResource(bitmapRsc, resourceID);
  bitmap = (BitmapType *)MemHandleLock(resH);
  WinDrawBitmap(bitmap, x + 1, y + 1);
  MemHandleUnlock(resH);
  DmReleaseResource(resH);
}

// Draw the cell with given index if it is currently visible
void CellDrawAbsolute(UInt16 xIx, UInt16 yIx)
{
  if ((xIx < board->width) && (yIx < board->height) &&
      (yIx >= currVPos) && (yIx < (currVPos + kVertCells)) &&
      (xIx >= currHPos) && (xIx < (currHPos + kHorizCells))) {
    CellDraw((xIx-currHPos)*kCellSize, (yIx-currVPos)*kCellSize, 
             xIx, yIx);
  }
}


static void BoardBoundaryDraw(FormPtr form, UInt16 gadgetID)
{
  RectangleType bounds;
  Coord x1, y1, x2, y2;

  FrmGetObjectBounds(form, FrmGetObjectIndex(form, BoardGadget), &bounds);

  bounds.extent.x = 1 + (MIN(kHorizCells, board->width)*kCellSize);
  bounds.extent.y = 1 + (MIN(kVertCells, board->height)*kCellSize);

  // Draw the S line which is not part of any cell
  x1 = bounds.topLeft.x;
  x2 = bounds.topLeft.x + bounds.extent.x - 1;
  y1 = bounds.topLeft.y + bounds.extent.y - 1;
  y2 = y1;
  WinDrawLine(x1,y1,x2,y2);

  // Draw the E line which is not part of any cell
  x1 = bounds.topLeft.x + bounds.extent.x - 1;
  x2 = x1;
  y1 = bounds.topLeft.y;
  y2 = bounds.topLeft.y + bounds.extent.y - 1;
  WinDrawLine(x1,y1,x2,y2);
}

void BoardGadgetDraw(FormPtr form, UInt16 gadgetID)
{
  UInt16 ii, jj;

  // Determine the offset of the main gadget if we haven't already
  if (!foundMainOffsets) {
    RectangleType bounds;
    FrmGetObjectBounds(form, FrmGetObjectIndex(form, BoardGadget), &bounds);
    mainDeltaX = bounds.topLeft.x; 
    mainDeltaY = bounds.topLeft.y;
    foundMainOffsets = true;
  }
  
  // Draw each of the (visible) cells
  for (ii=0; ii<kHorizCells; ii++) {
    for (jj=0; jj<kVertCells; jj++) {
      CellDraw(ii*kCellSize, jj*kCellSize, currHPos+ii, currVPos+jj);
    }
  }
  // Draw the board boundary
  BoardBoundaryDraw(form, gadgetID);
  // Show the numbers 
  BoardShowNumbers(form);
}

static MemHandle resUp;
static BitmapType *bitmapUp;
static MemHandle resDown;
static BitmapType *bitmapDown;

void BoardGadgetInit()
{
  resDown = DmGetResource(bitmapRsc, PressedBitmap);
  bitmapDown = (BitmapType *)MemHandleLock(resDown);
  resUp = DmGetResource(bitmapRsc, NormalBitmap);
  bitmapUp = (BitmapType *)MemHandleLock(resUp);
}

void BoardGadgetFinit()
{
  MemHandleUnlock(resUp);
  DmReleaseResource(resUp);
  MemHandleUnlock(resDown);
  DmReleaseResource(resDown);
}

void BoardGadgetTap(FormPtr form, UInt16 gadgetID, EventPtr event)
{
  Int16 ii;
  UInt16 xIx, yIx;
  RectangleType bounds;
  Boolean penDown;
  Boolean wasInBounds = true;
  UInt8 cell;
  Boolean flagMode = false;
  Boolean deduceMode = false;
  Int16 exposeCount = 0;
  enum {None, Flag, Expose, Toggle2, Toggle3, Deduce} action;
  enum {OneCell, NbrCells, AllCells} display;

  // First, figure out which cell the penDown is in and its bounds
  if (event->screenX >= (mainDeltaX + (kHorizCells*kCellSize))) return;
  if (event->screenY >= (mainDeltaY + (kVertCells*kCellSize))) return;
  xIx = ((event->screenX - mainDeltaX) / kCellSize);
  yIx = ((event->screenY - mainDeltaY) / kCellSize);
  bounds.topLeft.x = mainDeltaX + (xIx * kCellSize) + 1;
  bounds.topLeft.y = mainDeltaY + (yIx * kCellSize) + 1;
  bounds.extent.x = kCellSize;
  bounds.extent.y = kCellSize;
  xIx += currHPos;
  yIx += currVPos;

  if (xIx >= board->width) return;
  if (yIx >= board->height) return;

  if (GAMEOVER(board)) {
    // Restart on any single click
    if (event->tapCount == 1) {
      InitBoard(board);
      CtlSetEnabled(FrmGetObjectPtr(form, FrmGetObjectIndex(form, DefuseButton)), true);
      BoardGadgetDraw(form, BoardGadget);
    }
    return;
  }

  // Figure out what type of click this is
  if (!defusePending) {
    if (KeyCurrentState() & keyBitPageDown) {
      flagMode = true;
    } else if (KeyCurrentState() & keyBitPageUp) {
      deduceMode = true;
    } else if (event->tapCount == 2) {
      deduceMode = true;
    }
  }
 
  // If this is a normal click on a normal cell, then display the cell 
  // differently while the pen is down 
  cell = CELLVALUE(board, xIx, yIx);
  if (!flagMode && !deduceMode && !FLAGGED(cell) && !REVEALED(cell) && !QUESTION(cell)) {
    Boolean downShowing;

    // Do highlighting until pen up
    WinDrawBitmap(bitmapDown, bounds.topLeft.x, bounds.topLeft.y);
    downShowing = true;
    do {
      Boolean nowInBounds;
      Int16 x, y;
      
      PenGetPoint(&x, &y, &penDown);
      nowInBounds = RctPtInRectangle(x, y, &bounds);
      if (nowInBounds != wasInBounds) {
        if (downShowing) {
          WinDrawBitmap(bitmapUp, bounds.topLeft.x, bounds.topLeft.y);
          downShowing = false;
        } else {
          WinDrawBitmap(bitmapDown, bounds.topLeft.x, bounds.topLeft.y);
          downShowing = true;
        }
        wasInBounds = nowInBounds;
      }
    } while (penDown);
  }
  //DEBUG((T, "value %01x at cell (%d, %d)", cell, xIx, yIx));
  
  // Figure out what action we take on the board
  action = None;
  if (wasInBounds) {
    // The user has selected a cell
    if (!GAMESTARTED(board)) {
      board->gameStatus |= GAMESTARTEDMASK;
      assert(board->timeTaken == 0);
      lastTimeCheck = TimGetTicks();
    }
    if (defusePending) {
      if (!REVEALED(cell)) {
        if (MINE(cell)) {
          action = Flag;
        } else {
          action = Expose;
        }
        for (ii=0; ii<kDefuseTime; ii++) {
          UInt32 timeNow;
          board->timeTaken += SysTicksPerSecond();
          BoardShowNumbers(form);
          timeNow = TimGetTicks();
          // Ideally, it would be better not to hard loop, but I don't know a sleep() equivalent function
          while (TimGetTicks() < (timeNow + kNumberShowDelay));
        }
      }
      defusePending = false;
      FrmSetControlValue(form, FrmGetObjectIndex(form, DefuseButton), defusePending);
    } else if (flagMode) {
      if (board->preferences & DOQUESTIONMARKS) {
        action = Toggle3;
      } else {
        action = Toggle2;
      }
    } else if (deduceMode) {
      action = Deduce;
    } else {
      if (!FLAGGED(cell)) {
        action = Expose;
      }
    }
  }

  // Perform the action on the board, and figure out how many cells may be affected as we go
  display = OneCell;
  switch (action) {
    case None:
      break;
    case Flag:
      FlagCell(board, xIx, yIx);
      break;
    case Toggle2:
      Toggle2Cell(board, xIx, yIx);
      break;
    case Toggle3:
      Toggle3Cell(board, xIx, yIx);
      break;
    case Expose:
      ExposeCell(board, xIx, yIx, &exposeCount);
      display = (exposeCount<=1) ? OneCell : AllCells;
      break;
    case Deduce:
      DeduceCell(board, xIx, yIx, &exposeCount);
      display = (exposeCount<=1) ? NbrCells : AllCells;
      break;
  }
  if (GAMELOST(board)) {
    ShowAllMines(board);
    CtlSetEnabled(FrmGetObjectPtr(form, FrmGetObjectIndex(form, DefuseButton)), false);
    display = AllCells;
  }

  // Redisplay the parts of the board that need it
  switch (display) {
    case OneCell:
      CellDrawAbsolute(xIx, yIx);
      break;
    case NbrCells:
      // Redraw all the neighbouring cells
      CellDrawAbsolute(xIx-1, yIx-1);
      CellDrawAbsolute(xIx,   yIx-1);
      CellDrawAbsolute(xIx+1, yIx-1);
      CellDrawAbsolute(xIx-1, yIx);
      CellDrawAbsolute(xIx,   yIx);
      CellDrawAbsolute(xIx+1, yIx);
      CellDrawAbsolute(xIx-1, yIx+1);
      CellDrawAbsolute(xIx,   yIx+1);
      CellDrawAbsolute(xIx+1, yIx+1);
      break;
    case AllCells:
      BoardGadgetDraw(form, gadgetID);
      break;
  }
  // Update the count of mines still to flag
  BoardShowNumbers(form);
 
  // Deal with end of game
  if (GAMEWON(board)) {
    UInt16 alertID = WonAlert;
    UInt16 timeTaken;

    // Update the time taken before we set high scores and suchlike
    UInt32 timeNow = TimGetTicks();
    board->timeTaken += (timeNow - lastTimeCheck);
    lastTimeCheck = timeNow;
    timeTaken = board->timeTaken / SysTicksPerSecond();
    BoardShowNumbers(form);

    if ((board->width == kSmallWidth) && (board->height == kSmallHeight)) {
      if ((board->totalMines == kSmallEasy) && (timeTaken < board->smallEasyTime)) {
        alertID = HighScoreAlert;
        board->smallEasyTime = timeTaken;
      } else if ((board->totalMines == kSmallMedium) && (timeTaken < board->smallMediumTime)) {
        alertID = HighScoreAlert;
        board->smallMediumTime = timeTaken;
      } else if ((board->totalMines == kSmallHard) && (timeTaken < board->smallHardTime)) {
        alertID = HighScoreAlert;
        board->smallHardTime = timeTaken;
      }
    } else if ((board->width == kMediumWidth) && (board->height == kMediumHeight)) {
      if ((board->totalMines == kMediumEasy) && (timeTaken < board->mediumEasyTime)) {
        alertID = HighScoreAlert;
        board->mediumEasyTime = timeTaken;
      } else if ((board->totalMines == kMediumMedium) && (timeTaken < board->mediumMediumTime)) {
        alertID = HighScoreAlert;
        board->mediumMediumTime = timeTaken;
      } else if ((board->totalMines == kMediumHard) && (timeTaken < board->mediumHardTime)) {
        alertID = HighScoreAlert;
        board->mediumHardTime = timeTaken;
      }
    } else if ((board->width == kLargeWidth) && (board->height == kLargeHeight)) {
      if ((board->totalMines == kLargeEasy) && (timeTaken < board->largeEasyTime)) {
        alertID = HighScoreAlert;
        board->largeEasyTime = timeTaken;
      } else if ((board->totalMines == kLargeMedium) && (timeTaken < board->largeMediumTime)) {
        alertID = HighScoreAlert;
        board->largeMediumTime = timeTaken;
      } else if ((board->totalMines == kLargeHard) && (timeTaken < board->largeHardTime)) {
        alertID = HighScoreAlert;
        board->largeHardTime = timeTaken;
      }
    }
    (void) FrmAlert(alertID);
    CtlSetEnabled(FrmGetObjectPtr(form, FrmGetObjectIndex(form, DefuseButton)), false);
  } 
}
